// @flow
import panable, { type PanMoveEvent } from '../Utils/PixiSimpleGesture/pan';
import transformRect from '../Utils/TransformRect';
import * as PIXI from 'pixi.js-legacy';
import { type ScreenType } from '../UI/Responsive/ScreenTypeMeasurer';
import InstancesSelection from './InstancesSelection';
import {
  type ResizeGrabbingLocation,
  resizeGrabbingLocationValues,
  resizeGrabbingRelativePositions,
  canMoveOnX,
  canMoveOnY,
} from './InstancesResizer';
import { type InstanceMeasurer } from './InstancesRenderer';
import Rectangle from '../Utils/Rectangle';
import KeyboardShortcuts from '../UI/KeyboardShortcuts';

type Props = {|
  instancesSelection: InstancesSelection,
  instanceMeasurer: InstanceMeasurer,
  onResize: (
    deltaX: number,
    deltaY: number,
    grabbingLocation: ResizeGrabbingLocation
  ) => void,
  onResizeEnd: () => void,
  onRotate: (number, number) => void,
  onRotateEnd: () => void,
  toCanvasCoordinates: (x: number, y: number) => [number, number],
  screenType: ScreenType,
  keyboardShortcuts: KeyboardShortcuts,
  onPanMove: (deltaX: number, deltaY: number, x: number, y: number) => void,
  onPanEnd: () => void,
  getFillColor: (isLocked: boolean) => {| color: number, alpha: number |},
|};

const getButtonSizes = (screenType: ScreenType) => {
  if (screenType === 'touch') {
    return {
      bigButtonSize: 14,
      smallButtonSize: 12,
      buttonPadding: 5,
      hitAreaPadding: 12,
    };
  }

  return {
    bigButtonSize: 10,
    smallButtonSize: 8,
    buttonPadding: 5,
    hitAreaPadding: 5,
  };
};

const RECTANGLE_BUTTON_SHAPE = 0;
const CIRCLE_BUTTON_SHAPE = 1;

const resizeGrabbingIconNames = {
  TopLeft: 'nwse-resize',
  BottomLeft: 'nesw-resize',
  BottomRight: 'nwse-resize',
  TopRight: 'nesw-resize',
  Top: 'ns-resize',
  Left: 'ew-resize',
  Bottom: 'ns-resize',
  Right: 'ew-resize',
};

/**
 * Render selection rectangle for selected instances.
 */
export default class SelectedInstances {
  instancesSelection: InstancesSelection;
  instanceMeasurer: InstanceMeasurer;
  onResize: (
    deltaX: number,
    deltaY: number,
    grabbingLocation: ResizeGrabbingLocation
  ) => void;
  onResizeEnd: () => void;
  onRotate: (number, number) => void;
  onRotateEnd: () => void;
  toCanvasCoordinates: (x: number, y: number) => [number, number];
  _screenType: ScreenType;
  keyboardShortcuts: KeyboardShortcuts;
  onPanMove: (deltaX: number, deltaY: number, x: number, y: number) => void;
  onPanEnd: () => void;
  getFillColor: (isLocked: boolean) => {| color: number, alpha: number |};

  pixiContainer = new PIXI.Container();
  rectanglesContainer = new PIXI.Container();
  selectedRectangles = [];
  resizeButtons: { [ResizeGrabbingLocation]: PIXI.Graphics } = {};
  rotateButton = new PIXI.Graphics();
  /**
   * When moving the view, a click on the buttons generated by this class
   * won't trigger a resize and interrupt the move of the view.
   * Similarly, when resizing an instance, if we start to move the view
   * (space key pressed), the resizing is interrupted to move the view.
   * This state is used to detect the change and call the "end" function
   * associated of the proper interaction (buttonInteraction or viewMoving).
   */
  _currentPanMovingGoal: null | 'buttonInteraction' | 'viewMoving' = null;

  constructor({
    instancesSelection,
    instanceMeasurer,
    onResize,
    onResizeEnd,
    onRotate,
    onRotateEnd,
    toCanvasCoordinates,
    screenType,
    keyboardShortcuts,
    onPanMove,
    onPanEnd,
    getFillColor,
  }: Props) {
    this.instanceMeasurer = instanceMeasurer;
    this.onResize = onResize;
    this.onResizeEnd = onResizeEnd;
    this.onRotate = onRotate;
    this.onRotateEnd = onRotateEnd;
    this.toCanvasCoordinates = toCanvasCoordinates;
    this.instancesSelection = instancesSelection;
    this._screenType = screenType;
    this.keyboardShortcuts = keyboardShortcuts;
    this.onPanMove = onPanMove;
    this.onPanEnd = onPanEnd;
    this.getFillColor = getFillColor;

    this.pixiContainer.addChild(this.rectanglesContainer);

    for (const resizeGrabbingLocation of resizeGrabbingLocationValues) {
      const resizeButton = new PIXI.Graphics();
      this.resizeButtons[resizeGrabbingLocation] = resizeButton;
      this._makeButton({
        objectButton: resizeButton,
        onMove: event => {
          this.onResize(event.deltaX, event.deltaY, resizeGrabbingLocation);
        },
        onEnd: () => {
          this.onResizeEnd();
        },
        onPanMove: event => {
          this.onPanMove(
            event.deltaX,
            event.deltaY,
            event.data.global.x,
            event.data.global.y
          );
        },
        onPanEnd: () => {
          this.onPanEnd();
        },
        cursor: resizeGrabbingIconNames[resizeGrabbingLocation],
      });
    }
    this._makeButton({
      objectButton: this.rotateButton,
      onMove: event => {
        this.onRotate(event.deltaX, event.deltaY);
      },
      onEnd: () => {
        this.onRotateEnd();
      },
      onPanMove: event => {
        this.onPanMove(
          event.deltaX,
          event.deltaY,
          event.data.global.x,
          event.data.global.y
        );
      },
      onPanEnd: () => {
        this.onPanEnd();
      },
      cursor: 'url("res/actions/rotate24_black.png"),auto',
    });
  }

  setScreenType(screenType: ScreenType) {
    this._screenType = screenType;
  }

  _makeButton({
    objectButton,
    onMove,
    onEnd,
    onPanMove,
    onPanEnd,
    cursor,
  }: {
    objectButton: PIXI.Graphics,
    onMove: (event: PanMoveEvent) => void,
    onEnd: () => void,
    onPanMove: (event: PanMoveEvent) => void,
    onPanEnd: () => void,
    cursor: string,
  }) {
    objectButton.eventMode = 'static';
    objectButton.buttonMode = true;
    objectButton.cursor = cursor;
    panable(objectButton);

    // change cursor style if space is pressed
    objectButton.addEventListener('mousemove', event => {
      if (this.keyboardShortcuts.shouldMoveView()) {
        objectButton.cursor = 'grab';
      } else {
        objectButton.cursor = cursor;
      }
    });

    objectButton.addEventListener('panmove', event => {
      if (this.keyboardShortcuts.shouldMoveView()) {
        if (this._currentPanMovingGoal === 'buttonInteraction') {
          onEnd();
        }
        onPanMove(event);
        this._currentPanMovingGoal = 'viewMoving';
      } else {
        onMove(event);
        this._currentPanMovingGoal = 'buttonInteraction';
      }
    });

    objectButton.addEventListener('panend', event => {
      if (this.keyboardShortcuts.shouldMoveView()) {
        onPanEnd();
      } else {
        onEnd();
      }
      this._currentPanMovingGoal = null;
    });
    this.pixiContainer.addChild(objectButton);
  }

  getPixiContainer() {
    return this.pixiContainer;
  }

  _renderButton(
    show: boolean,
    buttonObject: PIXI.Graphics,
    canvasPosition: [number, number],
    size: number,
    shape: 0 | 1,
    hitAreaPadding: number
  ) {
    buttonObject.clear();
    if (!show) {
      buttonObject.hitArea = new PIXI.Rectangle(0, 0, 0, 0);
      return;
    }

    buttonObject.beginFill(0xffffff);
    buttonObject.lineStyle(1, 0x6868e8, 1);
    buttonObject.fill.alpha = 0.9;
    if (shape === RECTANGLE_BUTTON_SHAPE) {
      buttonObject.drawRect(canvasPosition[0], canvasPosition[1], size, size);
    } else if (shape === CIRCLE_BUTTON_SHAPE) {
      buttonObject.drawCircle(
        canvasPosition[0] + size / 2,
        canvasPosition[1] + size / 2,
        size / 2
      );
    }

    buttonObject.endFill();
    buttonObject.hitArea = new PIXI.Rectangle(
      canvasPosition[0] - hitAreaPadding,
      canvasPosition[1] - hitAreaPadding,
      size + hitAreaPadding * 2,
      size + hitAreaPadding * 2
    );
  }

  getSelectionAABB = (): Rectangle => {
    const selectionAABB = new Rectangle();
    const selection = this.instancesSelection.getSelectedInstances();
    let instanceRect = new Rectangle();
    for (var i = 0; i < selection.length; i++) {
      const instance = selection[i];

      instanceRect = this.instanceMeasurer.getInstanceAABB(
        instance,
        instanceRect
      );
      if (i === 0) {
        selectionAABB.setRectangle(instanceRect);
      } else {
        selectionAABB.union(instanceRect);
      }
    }
    return selectionAABB;
  };

  render() {
    const {
      bigButtonSize,
      smallButtonSize,
      buttonPadding,
      hitAreaPadding,
    } = getButtonSizes(this._screenType);
    const selection = this.instancesSelection.getSelectedInstances();
    let x1 = 0;
    let y1 = 0;
    let x2 = 0;
    let y2 = 0;
    let initialised = false;

    // Update the selection rectangle of each instance.
    for (var i = 0; i < selection.length; i++) {
      if (this.selectedRectangles.length === i) {
        const newRectangle = new PIXI.Graphics();
        newRectangle.hitArea = new PIXI.Rectangle(0, 0, 0, 0);
        this.selectedRectangles.push(newRectangle);
        this.rectanglesContainer.addChild(newRectangle);
      }

      const instance = selection[i];
      const instanceRect = this.instanceMeasurer.getInstanceAABB(
        instance,
        new Rectangle()
      );
      const selectionRectangle = transformRect(
        this.toCanvasCoordinates,
        instanceRect
      );

      this.selectedRectangles[i].clear();
      const { color, alpha } = this.getFillColor(instance.isLocked());
      this.selectedRectangles[i].beginFill(color, alpha);
      this.selectedRectangles[i].lineStyle(1, color, 1);
      this.selectedRectangles[i].fill.alpha = 0.3;
      this.selectedRectangles[i].alpha = 0.8;
      this.selectedRectangles[i].drawRect(
        selectionRectangle.left,
        selectionRectangle.top,
        selectionRectangle.width(),
        selectionRectangle.height()
      );
      this.selectedRectangles[i].endFill();

      if (instance.isLocked()) {
        continue;
      }
      if (!initialised) {
        x1 = instanceRect.left;
        y1 = instanceRect.top;
        x2 = instanceRect.left + instanceRect.width();
        y2 = instanceRect.top + instanceRect.height();
        initialised = true;
      } else {
        if (instanceRect.left < x1) x1 = instanceRect.left;
        if (instanceRect.top < y1) y1 = instanceRect.top;
        if (instanceRect.left + instanceRect.width() > x2)
          x2 = instanceRect.left + instanceRect.width();
        if (instanceRect.top + instanceRect.height() > y2)
          y2 = instanceRect.top + instanceRect.height();
      }
    }

    while (this.selectedRectangles.length > selection.length) {
      this.rectanglesContainer.removeChild(this.selectedRectangles.pop());
    }

    // If there are no unlocked instances, hide the resize buttons.
    const show =
      selection.filter(instance => !instance.isLocked()).length !== 0;

    // Position the resize buttons.
    for (const grabbingLocation of resizeGrabbingLocationValues) {
      const resizeButton = this.resizeButtons[grabbingLocation];
      const relativePositions =
        resizeGrabbingRelativePositions[grabbingLocation];

      const useBigButton =
        canMoveOnX(grabbingLocation) && canMoveOnY(grabbingLocation);
      const buttonSize = useBigButton ? bigButtonSize : smallButtonSize;
      const padding = buttonPadding + buttonSize / 2;
      const resizeButtonPos = this.toCanvasCoordinates(
        x1 + relativePositions[0] * (x2 - x1),
        y1 + relativePositions[1] * (y2 - y1)
      );
      resizeButtonPos[0] +=
        -buttonSize / 2 + padding * Math.sign(relativePositions[0] - 0.5);
      resizeButtonPos[1] +=
        -buttonSize / 2 + padding * Math.sign(relativePositions[1] - 0.5);

      this._renderButton(
        show,
        resizeButton,
        resizeButtonPos,
        buttonSize,
        RECTANGLE_BUTTON_SHAPE,
        hitAreaPadding
      );
    }

    const rotateButtonPos = this.toCanvasCoordinates(x1 + (x2 - x1) / 2, y1);
    rotateButtonPos[0] -= smallButtonSize / 2;
    rotateButtonPos[1] -= buttonPadding * 8;

    this._renderButton(
      show,
      this.rotateButton,
      rotateButtonPos,
      smallButtonSize,
      CIRCLE_BUTTON_SHAPE,
      hitAreaPadding
    );
  }
}
